#!/usr/bin/env python
'''
    This is a simple script where it all comes together.
    You can make you own if you want but this one is rather simple
'''

import ConfigParser
import os
import shutil
import sys
import tarfile
import tempfile

# make it easier to test this from the source archive
parent, bindir = os.path.split(os.path.dirname(os.path.abspath(sys.argv[0])))
if os.path.exists(os.path.join(parent, 'rpmpatch')):
    sys.path.insert(0, parent)

from rpmpatch.SourcePackage import SourcePackage
from rpmpatch.SpecFile import SpecFile

def rpmpatch(configdir, srpm, changelog_user, verbose=False):
    '''
        Read the config, perform the steps, build the package
    '''
    package = SourcePackage(srpm)
    name = package.packageName()
    version = package.packageVersion()
    cfgfilename = configdir + '/' + name + '.ini'
    if not os.path.isfile(cfgfilename):
        otherspot = configdir + '/' + name + '/' + name + '.ini'
        if not os.path.isfile(otherspot):
            raise RuntimeError('No config file found for ' + srpm)
        else:
            cfgfilename = configdir + '/' + name + '/' + name + '.ini'

    config = ConfigParser.SafeConfigParser()
    config.read(cfgfilename)

    configdir = os.path.abspath(os.path.dirname(cfgfilename))

    cfg_file = {}
    for section in config.sections():
        cfg_file[section] = {}
        for entry in config.options(section):
            value = config.get(section, entry)
            cfg_file[section.lower()][entry] = value

    # empty is ok, but we do need these
    if not cfg_file.has_key('program'):
        cfg_file['program'] = {}
    if not cfg_file.has_key('define'):
        cfg_file['define'] = []

    # need changelog user or else
    if changelog_user == None:
        if cfg_file['program'].has_key('changelog_user'):
            changelog_user = cfg_file['program']['changelog_user']
        else:
            print >> sys.stderr, '\nMissing changelog user, see --help or config file'
            sys.exit(1)

    if verbose:
        print '# working on ' + srpm

    # don't bother installing until we have a config file
    package.install()

    specfile = SpecFile(package.specfile, changelog_user)

    not_applicable = []
    if not cfg_file['program'].has_key('package_config'):
        cfg_file['program'] = False
    if not cfg_file['program'].has_key('package_unused'):
        cfg_file['program']['package_unused'] = False

    if not cfg_file['program'].has_key('compile'):
        cfg_file['program']['compile'] = False
    if not cfg_file['program'].has_key('build_targets'):
        cfg_file['program']['build_targets'] = None

    # get into sorted order
    sections = cfg_file.keys()
    sections.sort()

    # do all specfile work
    for section in sections:
        if section.startswith('spec'):
            if not cfg_file[section]['diff'].startswith('/'):
                cfg_file[section]['diff'] = configdir + '/' + cfg_file[section]['diff']
            result = spec_patch(cfg_file[section], version, specfile)
            if result == NotImplemented:
                not_applicable.append(cfg_file[section]['diff'])
            elif verbose:
                print '## applied ' + cfg_file[section]['diff']

    # do all regex work
    for section in sections:
        if section.startswith('re'):
            result = run_re(cfg_file[section], version, specfile)
            if result == NotImplemented:
                pass
            elif verbose:
                print '## ran regex ' + cfg_file[section]['match'] + '=>' + cfg_file[section]['replace']

    # do all patch work
    for section in sections:
        if section.startswith('patch'):
            if not cfg_file[section]['patch'].startswith('/'):
                cfg_file[section]['patch'] = configdir + '/' + cfg_file[section]['patch']
            result = patches(cfg_file[section], section, version, specfile)
            if result == NotImplemented:
                not_applicable.append(cfg_file[section]['patch'])
            elif verbose:
                print '## ' + cfg_file[section]['method'] + ' patch ' + result[1]

    # do all source work
    for section in sections:
        if section.startswith('source'):
            if not cfg_file[section]['source'].startswith('/'):
                cfg_file[section]['source'] = configdir + '/' + cfg_file[section]['source']
            result = sources(cfg_file[section], section, version, specfile)
            if result == NotImplemented:
                not_applicable.append(cfg_file[section]['source'])
            elif verbose:
                print '## added source ' + result[1]

    if cfg_file['program']['package_config']:
        fake_source = {}
        fake_source['method'] = 'add'
        fake_source['source'] = cfgfilename
        fake_source['changelog'] = 'Config file for automated patch script'
        sources(fake_source, 'program', version, specfile)
        if verbose:
            print '## added config file'

    if not_applicable:
        if cfg_file['program']['package_unused']:
            tararchive = name + '-extras-python-rpmpatch.tar.bz2'
            tempdir = tempfile.mkdtemp()
            tar = tarfile.open(tempdir + '/' + tararchive, 'w:bz2')
            os.chdir(tempdir)
            for filepath in not_applicable:
                shutil.copy2(filepath, tempdir)
                filename = os.path.basename(filepath)
                tar.add(filename)
            tar.close()

            fake_source = {}
            fake_source['method'] = 'add'
            fake_source['source'] = tararchive
            fake_source['changelog'] = 'Un-used items from automated patch script'

            sources(fake_source, 'program', version, specfile)

            os.chdir(tempfile.gettempdir())
            # done with this, clean it up
            shutil.rmtree(tempdir)
            if verbose:
                print '## added un-used items'

    if verbose:
        print '## building rpm'
    built = specfile.build(cfg_file['define'], cfg_file['program']['compile'],
                           cfg_file['program']['build_targets'])
    return built

def spec_patch(config, version, specfile):
    '''
        Apply a patch to the specfile
    '''
    if config.has_key('on_version'):
        if config['on_version'].startswith('['):
            config['on_version'] = eval(config['on_version'])
        else:
            config['on_version'] = [ config['on_version'] ]

        if version not in config['on_version']:
            return NotImplemented

    changelog = config['changelog']
    patch = config['diff']
    return specfile.apply_specfile_diff(patch, changelog)

def run_re(config, version, specfile):
    '''
        Run the listed regex against the specfile
    '''
    if config.has_key('on_version'):
        if config['on_version'].startswith('['):
            config['on_version'] = eval(config['on_version'])
        else:
            config['on_version'] = [ config['on_version'] ]

        if version not in config['on_version']:
            return NotImplemented

    changelog = config['changelog']
    find = config['match']
    replace = config['replace']
    return specfile.run_re(find, replace, changelog)

def patches(config, section, version, specfile):
    '''
        Add or remove the patch listed
    '''
    if config.has_key('on_version'):
        if config['on_version'].startswith('['):
            config['on_version'] = eval(config['on_version'])
        else:
            config['on_version'] = [ config['on_version'] ]

        if version not in config['on_version']:
            return NotImplemented

    method = config['method']
    changelog = config['changelog']
    if method.lower() == 'add':
        name = config['patch']
        stripe = config['stripe']
        if 'num' in config.keys():
            num = config['num']
        else:
            num = None
        return specfile.add_patch(name, stripe, num, changelog)

    if method.lower() == 'del':
        if 'num' in config.keys():
            num = config['num']
        else:
            num = None

        if 'patch' in config.keys():
            name = os.path.basename(config['patch'])
        else:
            name = None

        if num == name:
            raise RuntimeError('Bad ' + section + ' in config file')

        return specfile.rm_patch(name, num, changelog)

def sources(config, section, version, specfile):
    '''
        Add a source to the specfile
    '''
    if config.has_key('on_version'):
        if config['on_version'].startswith('['):
            config['on_version'] = eval(config['on_version'])
        else:
            config['on_version'] = [ config['on_version'] ]

        if version not in config['on_version']:
            return NotImplemented

    changelog = config['changelog']
    method = config['method']
    if method.lower() == 'add':
        if 'num' in config.keys():
            num = config['num']
        else:
            num = None
        thisfile = config['source']
        return specfile.add_source(thisfile, num, changelog)
    else:
        raise RuntimeError('Bad ' + section + ' in config file')

def parsesrpms(configdir, srpms, changelog_user, verbose=False):
    '''
        Loop through the rpms, possible use of threading here later
    '''
    result = []
    for srpm in srpms:
        result.append(rpmpatch(configdir, srpm, changelog_user, verbose))

    for item in result:
        for element in item:
            print "Wrote: " + element

if __name__ == "__main__":

    import optparse
    import os
    import rpm
    import sys

    # don't fix epilog white space, useful for manual formating control
    optparse.OptionParser.format_epilog = lambda self, formatter: self.epilog

    DEF_CONF = rpm.expandMacro('%_topdir') + '/CUSTOM_PATCHES'

    DESC = '''This program will take in a config directory along with source
              packages and then customize them for you based on config files.'''

    EPILOG = '\nExamples:\n\n'
    EPILOG = EPILOG + ' ' + sys.argv[0]
    EPILOG = EPILOG + ' my.src.rpm \n'
    EPILOG = EPILOG + ' ' + sys.argv[0]
    EPILOG = EPILOG + ' --configdir=/etc/myconf.d/  my.src.rpm\n'
    EPILOG = EPILOG + ' ' + sys.argv[0]
    EPILOG = EPILOG + ' --configdir=/etc/myconf.d/  my.src.rpm'
    EPILOG = EPILOG + ' --changelog_user=me\n'
    EPILOG = EPILOG + ' ' + sys.argv[0]
    EPILOG = EPILOG + ' --configdir=/etc/myconf.d/  my.src.rpm --verbose\n'
    EPILOG = EPILOG + '\n' + '#' * 78 + '\n'

    SAMPLE_CONFIG = '''

Things are run in groups organized by their section names.
Each section is run alphabeticaly, but the sections are run in this order:

spec-> re -> patch -> source

This way your specfile adjustments and regex matches don't need to contend
with your other modifications

-------------------------------------------------
Example Config:
-------------------------------------------------
[program]
# you can control behavior here

# you can override this on the command line
changelog_user = me <my@addr.com>

# if you set this to 'True' the ini file used
# to configure this program will be added to the
# source rpm
package_config = True

# if you set this to 'True' any patches or sources
# that were not applied due to version restrictions
# will be placed in a tar archive called
# %{name}-extras-python-rpmpatch.tar.bz2
package_unused = True

# if set to 'True' then rpmbuild -bb will be called
# automatically rather than just rpmbuild -bs
compile = False

# which arches should this rpm be built for
# if this is defined, when a compile is requested
# the items here will be passed to rpmbuild as 
# a --target=<> 
build_targets = [ 'i386', 'i686' ]

# anything you want set for rpmbuild, mark here
[define]
dist = .el6
macro = value

# ----------------------------------------------------
# The patch sections are either for adding or removing.
[patch1]
# example for adding a patch
method = add

# DO NOT PUT THESE IN rpmbuild/SOURCES they will be copied in there for you
patch = /path/to/patch

# the arg to -p, ie 1 for -p1
stripe = 1

# you can leave the 'num' line off or set it if you want
num = 200

# on package version, this should be a list of versions
# for which this is performed.  The sort of thing you
# get from the %{version} macro
# If undefined, we assume you want everyone to get this
on_version = [ '1.2' ]

# you must have a changelog entry
changelog = I'm adding this patch to resolve issue and fix bug

# ----------------------------------------------------
[patch2]
# example for removing a patch
method = del

# you must either specify patch name or patch num
# your choice
patch = mypatch.patch
num = 12

# on package version, this should be a list of versions
# for which this is performed.  The sort of thing you
# get from the %{version} macro
# If undefined, we assume you want everyone to get this
on_version = [ '1.2' ]

# you must have a changelog entry
changelog = I'm removing this patch because it causes problem

# ----------------------------------------------------
# The 'source' sections are for adding sources right now
# Removing sources requires more complex work (like a patch)
# You probably still need a patch to do something with your newly added source
[source1]
# example for adding a source
method = add

# DO NOT PUT THESE IN rpmbuild/SOURCES they will be copied in there for you
source = /path/to/source/file

# you can leave the 'num' line off or set it if you want
num = 200

# on package version, this should be a list of versions
# for which this is performed.  The sort of thing you
# get from the %{version} macro
# If undefined, we assume you want everyone to get this
on_version = [ '1.2' ]

# you must have a changelog entry
changelog = An extra thingy

# ----------------------------------------------------
# The 'spec' sections are for applying patches to the specfile itself
# when you've added a new source or want to remove a source file
# this is how you can do it.
[spec1]
# The patch will be added as a 'Source' so you can review it later
diff = /path/to/patch

# on package version, this should be a list of versions
# for which this is performed.  The sort of thing you
# get from the %{version} macro
# If undefined, we assume you want everyone to get this
on_version = [ '1.2' ]

# you must have a changelog entry
changelog = I'm making some important changes such as 

# ----------------------------------------------------
# the 're' sections are for running any random regex you want
# against the specfile.  It isn't as focused as a patch, but that
# can be a benefit
[re1]

# set your expression up here
match = .
replace = L

# on package version, this should be a list of versions
# for which this is performed.  The sort of thing you
# get from the %{version} macro
# If undefined, we assume you want everyone to get this
on_version = [ '1.2' ]

# you must have a changelog entry
changelog = I changed everything to L for some reason

'''

    PARSER = optparse.OptionParser(description=DESC, epilog=EPILOG)

    PARSER.add_option('--configdir', metavar="<"+DEF_CONF+">", default=DEF_CONF,
                      help='''Path to a directory containing ini style configs
                              for the rpms you wish to patch.  The config
                              filenames should be in the format %{name}.ini
                              or %{name}/%{name}.ini your choice.
                              For example, http-2.2-2.el5.i386.rpm would have
                              a config file named 'httpd.ini' within this
                              directory or a directory named 'httpd' with
                              a config file named 'httpd.ini' within.
                              The default is within your 'rpmbuild' location''')

    PARSER.add_option('--changelog_user', metavar="<USER>", default=None,
                      help='''Name of user who gets credit for these changes.
                              You can define this in the config if you want.
                              The command line will override this value if set.                               (OPTIONAL)''')

    PARSER.add_option('--sampleconfig', default=False, action='store_true',
                      help='''Show a sample config file''')

    PARSER.add_option('--verbose', default=False, action='store_true',
                      help='''Print action information''')

    (OPTIONS, ARGS) = PARSER.parse_args()

    if OPTIONS.sampleconfig:
        print SAMPLE_CONFIG
        sys.exit(1)

    if not os.path.isdir(OPTIONS.configdir):
        PARSER.print_help()
        print ''
        raise IOError('No such directory: ' + OPTIONS.configdir)

    if len(ARGS) < 1:
        PARSER.print_help()
        print >> sys.stderr, '\nMissing srpm(s), see help'

    for THEFILE in ARGS:
        if not os.path.isfile(THEFILE):
            raise IOError('No such file: ' + OPTIONS.configdir)

    parsesrpms(OPTIONS.configdir, ARGS, OPTIONS.changelog_user, OPTIONS.verbose)
